package com.softslate.commerce.businessobjects.order;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.beanutils.ConvertUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.BaseBusinessProcessor;
import com.softslate.commerce.businessobjects.customer.Customer;
import com.softslate.commerce.businessobjects.events.ChangedOrderEvent;
import com.softslate.commerce.businessobjects.events.ChangingOrderEvent;
import com.softslate.commerce.daos.order.OrderDAO;
import com.softslate.commerce.daos.order.OrderGatewayDAO;

/**
 * Processes business logic affecting orders and other related objects.
 * <code>BasicOrderProcessor</code> is the default implementation of the
 * <code>OrderProcessor</code> interface for the application.
 * 
 * @author David Tobey
 * @see com.softslate.commerce.businessobjects.order.OrderProcessor
 */
public class BasicOrderProcessor extends BaseBusinessProcessor implements OrderProcessor {

	static Log log = LogFactory.getLog(BasicOrderProcessor.class);

	public StatesAndCountries loadStatesAndCountries() throws Exception {
		StatesAndCountries statesAndCountries = (StatesAndCountries) getBusinessObjectFactory().createObject(
				"statesAndCountriesImplementer");
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		statesAndCountries = orderGatewayDAO.loadStatesAndCountries(statesAndCountries);
		Collection countries = statesAndCountries.getCountries();
		if (countries != null) {
			statesAndCountries.setCountryMap(new HashMap());
			Iterator i = countries.iterator();
			while (i.hasNext()) {
				Country c = (Country) i.next();
				statesAndCountries.getCountryMap().put(c.getCode(), c);
			}
		}
		Collection states = statesAndCountries.getStates();
		if (states != null) {
			statesAndCountries.setStateMap(new HashMap());
			Iterator i = states.iterator();
			while (i.hasNext()) {
				State s = (State) i.next();
				statesAndCountries.getStateMap().put(s.getCode(), s);
			}
		}
		return statesAndCountries;
	}

	public Collection loadDashboardStatistics() throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadDashboardStatistics();
	}

	public Collection loadIncompleteOrdersStatistics() throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadIncompleteOrdersStatistics();
	}

	public int deleteIncompleteOrders(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Deleting incomplete orders");
		OrderGatewayDAO dao =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return dao.deleteIncompleteOrders(true, parameters);
	}

	public Map loadOrdersAndCount(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Loading orders and count");
		OrderGatewayDAO gatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return gatewayDAO.loadOrdersAndCount(parameters);
	}

	public Map addOrder(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Adding Order");
		Order order =  getBusinessObjectFactory().createObject(Order.class);
		Map originalParameters = BeanUtils.describe(order);
		Map results = postChangingOrderEvent(order, parameters);
		BeanUtils.copyProperties(order, parameters);
		if (order.getCustomerID() != null && order.getCustomerID().intValue() != 0) {
			Customer customer =  getBusinessObjectFactory().createObject(Customer.class);
			customer.setCustomerID(order.getCustomerID().intValue());
			order.setCustomer(customer);
		}
		OrderDAO orderDAO =  getDaoFactory().createDAO(OrderDAO.class);
		orderDAO.setOrder(order);
		orderDAO.insertOrder(true);
		results.putAll(postChangedOrderEvent(order, originalParameters));
		results.put("newRecord", orderDAO.getOrder());
		return results;
	}

	public Order loadOrderFromID(Map parameters) throws Exception {
		return loadOrderFromID((Integer) ConvertUtils.convert(parameters.get("orderID"), Integer.class));
	}

	public Order loadOrderFromID(int orderID) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Loading order from id");
		Order order =  getBusinessObjectFactory().createObject(Order.class);
		OrderDAO dao =  getDaoFactory().createDAO(OrderDAO.class);
		order.setOrderID(orderID);
		dao.setOrder(order);
		dao.loadOrderFromID();
		return dao.getOrder();
	}

	public Order loadOrderFromNumber(Map parameters) throws Exception {
		return loadOrderFromNumber(new Integer((String) parameters.get("orderNumber")));
	}

	public Order loadOrderFromNumber(int orderNumber) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Loading order from orderNumber");
		Order order =  getBusinessObjectFactory().createObject(Order.class);
		OrderDAO dao =  getDaoFactory().createDAO(OrderDAO.class);
		order.setOrderNumber(orderNumber);
		dao.setOrder(order);
		dao.loadOrderFromNumber();
		return dao.getOrder();
	}

	public Map editOrder(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Editing Order");
		Order order = null;
		if (parameters.get("orderID") != null) {
			order = loadOrderFromID(parameters);
		}
		if (order == null && parameters.get("orderNumber") != null) {
			order = loadOrderFromNumber(parameters);
		}
		Map originalParameters = BeanUtils.describe(order);
		Map results = postChangingOrderEvent(order, parameters);
		BeanUtils.copyProperties(order, parameters);
		if (parameters.get("formCustomerID") != null && !parameters.get("formCustomerID").equals("")
				&& !parameters.get("formCustomerID").equals("0")) {
			int customerID = new Integer((String) parameters.get("formCustomerID"));
			Customer customer =  getBusinessObjectFactory().createObject(Customer.class);
			customer.setCustomerID(customerID);
			order.setCustomer(customer);
			order.setCustomerID(customerID);
		} else if (parameters.get("formCustomerID") != null
				&& (parameters.get("formCustomerID").equals("") || parameters.get("formCustomerID").equals("0"))) {
			order.setCustomer(null);
		}
		OrderDAO orderDAO =  getDaoFactory().createDAO(OrderDAO.class);
		orderDAO.setOrder(order);
		orderDAO.updateOrder(true);
		results.putAll(postChangedOrderEvent(order, originalParameters));
		results.put("updatedRecord", orderDAO.getOrder());
		return results;
	}

	public Map postChangingOrderEvent(Order order, Map parameters) throws Exception {
		Map results = new HashMap();
		ChangingOrderEvent event = (ChangingOrderEvent) getBusinessObjectFactory().createObject(
				"changingOrderEventImplementer");
		event.setBusinessProcessor(this);
		event.setOrder(order);
		event.setIncomingParameters(parameters);
		event.setResults(results);
		getEventBus().post(event);
		return event.getResults();
	}

	public Map postChangedOrderEvent(Order order, Map parameters) throws Exception {
		Map results = new HashMap();
		ChangedOrderEvent event = (ChangedOrderEvent) getBusinessObjectFactory().createObject(
				"changedOrderEventImplementer");
		event.setBusinessProcessor(this);
		event.setOrder(order);
		event.setOriginalParameters(parameters);
		event.setResults(results);
		getEventBus().post(event);
		return event.getResults();
	}

	public Map editDeleteOrders(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Editing/deleting multiple orders");
		OrderGatewayDAO dao =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		Map results = dao.editDeleteOrders(parameters, getBusinessObjectFactory());
		return results;
	}

	public void deleteOrder(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("orderNumber");
			log.info(getUser().getAdministrator().getUserName() + ": Deleting order: " + code);
		}
		Order order =  getBusinessObjectFactory().createObject(Order.class);
		BeanUtils.copyProperties(order, parameters);
		OrderDAO dao =  getDaoFactory().createDAO(OrderDAO.class);
		dao.setOrder(order);
		dao.deleteOrder(true);
		return;
	}

	public Order loadOrderDetails(int orderNumber) throws Exception {
		if (log.isDebugEnabled()) {
			log.debug("Getting order details for order number: " + orderNumber);
		}
		// Query for the order
		Order order =  getBusinessObjectFactory().createObject(Order.class);
		order.setOrderNumber(new Integer(orderNumber));
		OrderDAO orderDAO =  getDaoFactory().createDAO(OrderDAO.class);
		orderDAO.setOrder(order);
		// Load the order details
		orderDAO.loadOrderFromNumber();
		if (orderDAO.getOrder() != null) {
			orderDAO.getOrder().addFormatting(getUser().getLocale(), (String) getSettings().getValue("currencyCode"));
		}
		return orderDAO.getOrder();
	}

	public Order duplicateOrder(Order order) throws Exception {
		if (log.isInfoEnabled())
			log.info("Duplicating order");
		if (order != null) {
			Order dupeOrder =  getBusinessObjectFactory().createObject(Order.class);
			BeanUtils.copyProperties(dupeOrder, order);
			dupeOrder.setOrderID(0);
			dupeOrder.setOrderNumber(null);
			dupeOrder.setCompleted(null);
			// Global discounts
			if (order.getOrderDiscounts() != null) {
				dupeOrder.setOrderDiscounts(new LinkedHashSet());
				Iterator h = order.getOrderDiscounts().iterator();
				while (h.hasNext()) {
					OrderDiscount discount = (OrderDiscount) h.next();
					if (!discount.getItemsAppliesType().equals("[skus]")) {
						OrderDiscount dupeDiscount = (OrderDiscount) getBusinessObjectFactory().createObject(
								"orderDiscountImplementer");
						BeanUtils.copyProperties(dupeDiscount, discount);
						dupeDiscount.setOrderDiscountID(0);
						dupeDiscount.setOrderID(0);
						dupeDiscount.setOrder(dupeOrder);
						dupeOrder.getOrderDiscounts().add(dupeDiscount);
					}
				}
			}
			// Global shipping rules
			if (order.getOrderShippingRules() != null) {
				dupeOrder.setOrderShippingRules(new LinkedHashSet());
				Iterator h = order.getOrderShippingRules().iterator();
				while (h.hasNext()) {
					OrderShippingRule osr = (OrderShippingRule) h.next();
					if (osr.getOrderItem() == null) {
						OrderShippingRule dupeOSR = (OrderShippingRule) getBusinessObjectFactory().createObject(
								"orderShippingRuleImplementer");
						BeanUtils.copyProperties(dupeOSR, osr);
						dupeOSR.setOrderShippingRuleID(0);
						dupeOSR.setOrderID(0);
						dupeOSR.setOrder(dupeOrder);
						dupeOrder.getOrderShippingRules().add(dupeOSR);
					}
				}
			}
			if (order.getOrderDeliveryCollection() != null) {
				dupeOrder.setOrderDeliveryCollection(new LinkedHashSet());
				Iterator i = order.getOrderDeliveryCollection().iterator();
				while (i.hasNext()) {
					OrderDelivery od = (OrderDelivery) i.next();
					OrderDelivery dupeOD = (OrderDelivery) getBusinessObjectFactory().createObject(
							"orderDeliveryImplementer");
					BeanUtils.copyProperties(dupeOD, od);
					dupeOD.setOrderDeliveryID(0);
					dupeOD.setOrderID(0);
					dupeOD.setOrder(dupeOrder);
					dupeOD.setDelivered(null);
					dupeOrder.getOrderDeliveryCollection().add(dupeOD);
					if (od.getOrderItemCollection() != null) {
						dupeOD.setOrderItemCollection(new LinkedHashSet());
						Iterator j = od.getOrderItemCollection().iterator();
						while (j.hasNext()) {
							OrderItem oi = (OrderItem) j.next();
							if (!oi.getIsActive()) {
								continue;
							}
							OrderItem dupeOI = (OrderItem) getBusinessObjectFactory().createObject(
									"orderItemImplementer");
							BeanUtils.copyProperties(dupeOI, oi);
							dupeOI.setOrderItemID(0);
							dupeOI.setOrderID(0);
							dupeOI.setOrder(dupeOrder);
							dupeOI.setOrderDeliveryID(0);
							dupeOI.setOrderDelivery(dupeOD);
							dupeOI.setDiscounts(null);
							dupeOD.getOrderItemCollection().add(dupeOI);
							if (oi.getOrderItemAttributes() != null) {
								dupeOI.setOrderItemAttributes(new LinkedHashSet());
								Iterator k = oi.getOrderItemAttributes().iterator();
								while (k.hasNext()) {
									OrderItemAttribute oia = (OrderItemAttribute) k.next();
									OrderItemAttribute dupeOIA =  getBusinessObjectFactory()
											.createObject(OrderItemAttribute.class);
									BeanUtils.copyProperties(dupeOIA, oia);
									dupeOIA.setOrderItemAttributeID(0);
									dupeOIA.setOrderItemID(0);
									dupeOIA.setOrderItem(dupeOI);
									dupeOI.getOrderItemAttributes().add(dupeOIA);
								}
							}
							// SKU discounts
							if (oi.getOrderDiscounts() != null) {
								dupeOI.setOrderDiscounts(new LinkedHashSet());
								if (dupeOrder.getOrderDiscounts() == null) {
									dupeOrder.setOrderDiscounts(new LinkedHashSet());
								}
								Iterator l = oi.getOrderDiscounts().iterator();
								while (l.hasNext()) {
									OrderDiscount discount = (OrderDiscount) l.next();
									OrderDiscount dupeDiscount =  getBusinessObjectFactory()
											.createObject(OrderDiscount.class);
									BeanUtils.copyProperties(dupeDiscount, discount);
									dupeDiscount.setOrderDiscountID(0);
									dupeDiscount.setOrderItemID(null);
									dupeDiscount.setOrderItem(dupeOI);
									dupeDiscount.setOrderID(0);
									dupeDiscount.setOrder(dupeOrder);
									dupeOI.getOrderDiscounts().add(dupeDiscount);
									dupeOrder.getOrderDiscounts().add(dupeDiscount);
								}
							}
							// SKU shipping rules
							if (oi.getOrderShippingRules() != null) {
								dupeOI.setOrderShippingRules(new LinkedHashSet());
								if (dupeOrder.getOrderShippingRules() == null) {
									dupeOrder.setOrderShippingRules(new LinkedHashSet());
								}
								Iterator l = oi.getOrderShippingRules().iterator();
								while (l.hasNext()) {
									OrderShippingRule osr = (OrderShippingRule) l.next();
									OrderShippingRule dupeOSR =  getBusinessObjectFactory()
											.createObject(OrderShippingRule.class);
									BeanUtils.copyProperties(dupeOSR, osr);
									dupeOSR.setOrderShippingRuleID(0);
									dupeOSR.setOrderItemID(null);
									dupeOSR.setOrderItem(dupeOI);
									dupeOSR.setOrderID(0);
									dupeOSR.setOrder(dupeOrder);
									dupeOI.getOrderShippingRules().add(dupeOSR);
									dupeOrder.getOrderShippingRules().add(dupeOSR);
								}
							}
						}
					}
				}
			}
			dupeOrder.setPayments(null);
			dupeOrder.setOrderItemCollection(null);
			return dupeOrder;
		}
		return null;
	}

	public Map loadSavedCartFromToken(Map parameters) throws Exception {
		if (log.isInfoEnabled()) {
			log.info("Loading auto saved cart for token: " + parameters.get("autoSavedCartToken"));
		}

		// Query for the customer's order
		Order order =  getBusinessObjectFactory().createObject(Order.class);
		order.setAutoSavedCartToken((String) parameters.get("autoSavedCartToken"));
		OrderDAO orderDAO =  getDaoFactory().createDAO(OrderDAO.class);
		orderDAO.setOrder(order);

		// Load the order
		orderDAO.loadSavedCartFromToken();

		// Either duplicate the order or reproccess the items, depending on the
		// 'Refresh Carts When Customers Pick Up' setting.
		getUser().setOrder(null);
		OrderProcessor orderProcessor = (OrderProcessor) getBusinessObjectFactory().createObject(
				"orderProcessorImplementer");
		if (getSettings().getValue("refreshSavedCarts").equals("1")) {
			Order savedOrder = getUser().getOrder();
			Map results = orderProcessor.reprocessOrder(orderDAO.getOrder());
			if (results == null) {
				getUser().setOrder(savedOrder);
			}
			return results;
		} else {
			Order dupeOrder = orderProcessor.duplicateOrder(orderDAO.getOrder());
			dupeOrder.setStatus("Copied");
			getUser().setOrder(dupeOrder);
			orderDAO.setOrder(dupeOrder);
			orderDAO.insertOrder(true);
			return null;
		}
	}

	public Map reprocessOrder(Order order) throws Exception {
		if (log.isInfoEnabled())
			log.info("Reprocessing order now");
		Map parameters = new HashMap();
		String[] orderProductQuantities;
		String[] orderProductCodes;
		Map orderAttributes = new HashMap();
		if (order != null) {
			if (order.getOrderDeliveryCollection() != null) {
				Iterator i = order.getOrderDeliveryCollection().iterator();
				while (i.hasNext()) {
					OrderDelivery od = (OrderDelivery) i.next();
					if (od.getOrderItemCollection() != null && !od.getOrderItemCollection().isEmpty()) {
						orderProductQuantities = new String[od.getOrderItemCollection().size()];
						orderProductCodes = new String[od.getOrderItemCollection().size()];
						int count = 0;
						Iterator j = od.getOrderItemCollection().iterator();
						while (j.hasNext()) {
							OrderItem oi = (OrderItem) j.next();
							orderProductQuantities[count] = Integer.toString(oi.getQuantity());
							orderProductCodes[count] = oi.getCode();
							if (oi.getOrderItemAttributes() != null) {
								Iterator k = oi.getOrderItemAttributes().iterator();
								while (k.hasNext()) {
									OrderItemAttribute oia = (OrderItemAttribute) k.next();
									String[] oias = (String[]) orderAttributes.get(oi.getCode() + "|"
											+ oia.getAttributeCode());
									if (oias == null) {
										oias = new String[od.getOrderItemCollection().size()];
									}
									oias[count] = oia.getAttributeValue();
									orderAttributes.put(oi.getCode() + "|" + oia.getAttributeCode(), oias);
								}
							}
							count++;
						}
						if (log.isInfoEnabled())
							log.info("orderAttributes: " + orderAttributes);
						parameters.put("orderProductQuantities", orderProductQuantities);
						parameters.put("orderProductCodes", orderProductCodes);
						parameters.put("orderAttributes", orderAttributes);
						parameters.put("autoSavedCartToken", order.getAutoSavedCartToken());
						CartProcessor processor = (CartProcessor) getBusinessObjectFactory().createObject(
								"cartProcessorImplementer");
						return processor.processAddItems(parameters);
					}
				}
			}
		}
		return null;
	}

	public int loadMaxOrderID() throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			log.info(getUser().getAdministrator().getUserName() + ": Loading max order ID.");
		}
		OrderGatewayDAO gatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return gatewayDAO.loadMaxOrderID();
	}

	public Collection loadIncompleteOrdersWithPayments(Map parameters) throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadIncompleteOrdersWithPayments(parameters);
	}

	public Collection<Order> loadCompletedOrdersByDateRange(String startDate, String endDate,
			Collection<String> ignoreStatuses) throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadCompletedOrdersByDateRange(startDate, endDate, ignoreStatuses);
	}

	public Collection<Order> loadAbandonedCartOrders(String startDate, String endDate, String prevOrdersDate)
			throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadAbandonedCartOrders(startDate, endDate, prevOrdersDate);
	}

	public Collection<Order> loadCompletedOrdersByLastModifiedAndStatuses(String startDate, String endDate,
			Collection<String> currentStatuses) throws Exception {
		OrderGatewayDAO orderGatewayDAO =  getDaoFactory().createDAO(OrderGatewayDAO.class);
		return orderGatewayDAO.loadCompletedOrdersByLastModifiedAndStatuses(startDate, endDate, currentStatuses);
	}

	public Collection processSettings(Map parameters) throws Exception {
		if (log.isInfoEnabled())
			log.info("About to process order settings.");
		// Load the settings
		Order order = (Order) getBusinessObjectFactory().createObject(
				"orderImplementer");
		BeanUtils.copyProperties(order, parameters);
		OrderDAO orderDAO = (OrderDAO) getDaoFactory().createDAO(
				"orderDAOImplementer");
		orderDAO.setOrder(order);
		Collection settings = orderDAO.loadSettings();
		parameters.put("settings", settings);
		parameters.put("orderDAO", orderDAO);
		if (log.isDebugEnabled())
			log.debug("Determining new order settings to insert.");
		// Create a collection of order settings that are new and must be
		// inserted.
		Collection orderSettingsToInsert = new ArrayList();
		String[] codes = (String[]) parameters.get("code");
		String[] names = (String[]) parameters.get("name");
		String[] descriptions = (String[]) parameters.get("description");
		String[] values = (String[]) parameters.get("value");
		String[] valueTypes = (String[]) parameters.get("valueType");
		for (int i = 0; i < codes.length; i++) {
			boolean insert = true;
			Iterator j = settings.iterator();
			while (j.hasNext()) {
				OrderSetting setting = (OrderSetting) j.next();
				if (setting.getCode().equals(codes[i])) {
					if (log.isDebugEnabled())
						log.debug("Existing order setting: " + codes[i]);
					setting.setOrder(order);
					insert = false;
				}
			}
			if (insert) {
				if (log.isDebugEnabled())
					log.debug("New order setting: " + codes[i]);
				OrderSetting orderSetting = (OrderSetting) getBusinessObjectFactory()
						.createObject("orderSettingImplementer");
				orderSetting.setOrderID(Integer.parseInt((String) parameters
						.get("orderID")));
				orderSetting.setCode(codes[i]);
				orderSetting.setName(names[i]);
				orderSetting.setDescription(descriptions[i]);
				orderSetting.setValueType(valueTypes[i]);
				orderSetting.setType((String) parameters.get("type"));
				if (valueTypes[i].equals("integerValue")
						|| valueTypes[i].equals("booleanValue")) {
					orderSetting.setIntegerValue(Integer.parseInt(values[i]));
				} else if (valueTypes[i].equals("smallValue")) {
					orderSetting.setSmallValue(values[i]);
				} else if (valueTypes[i].equals("mediumValue")) {
					orderSetting.setMediumValue(values[i]);
				} else if (valueTypes[i].equals("textValue")) {
					orderSetting.setTextValue(values[i]);
				}
				orderSetting.setOrder(order);
				orderSettingsToInsert.add(orderSetting);
			}
		}
		parameters.put("orderSettingsToInsert", orderSettingsToInsert);
		return orderDAO.processSettings(true, parameters);
	}

}
