package com.softslate.commerce.businessobjects.payment;

import com.softslate.commerce.daos.payment.PaymentDAO;

import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.payment.authorizeNet.AuthorizeNetAIMConstants;

/**
 * Processes a payment in the store for a given user's cart. This implementation
 * processes a request through Authorize.net's AIM integration method.
 * <p>
 * If the transaction is declined a result code of 10000 is returned, if it is
 * successful a result code of 0 is returned, and otherwise the result code sent
 * back from Authorize.net is returned.
 * 
 * @author Justin Casp, David Tobey
 * @see com.softslate.commerce.daos.payment.PaymentDAO
 */
public class AuthorizeNetProcessor extends BasePaymentProcessor implements
		PaymentProcessor {

	static Log log = LogFactory.getLog(AuthorizeNetProcessor.class);

	// Authorize.Net response codes
	public static final int RESPONSE_APPROVED = 1;

	public static final int RESPONSE_DECLINED = 2;

	public static final int RESPONSE_ERROR = 3;

	protected Map gatewayPostParams = null;

	public Map initializePostParams(Map parameters) throws Exception {

		if (gatewayPostParams == null)
			gatewayPostParams = new HashMap();

		// merchant account fields
		gatewayPostParams.put("x_login", getSettings().getValue(
				"authorizeNetLoginID"));
		gatewayPostParams.put("x_tran_key", getSettings().getValue(
				"authorizeNetTxKey"));
		gatewayPostParams.put("x_currency_code", getSettings().getValue(
				"currencyCode"));

		// test mode
		String testMode = (String) getSettings().getValue(
				"authorizeNetTestRequest");
		if (testMode != null && testMode.equals("1")) {
			gatewayPostParams.put("x_test_request", "TRUE");
		}

		// email control fields
		String emailCust = (String) getSettings().getValue(
				"authorizeNetEmailCustomer");
		if (emailCust != null && emailCust.equals("1")) {
			gatewayPostParams.put("x_email_customer", "TRUE");
		} else {
			gatewayPostParams.put("x_email_customer", "FALSE");
		}
		gatewayPostParams.put("x_merchant_email", getSettings().getValue(
				"authorizeNetEmailMerchant"));

		// gateway response fields
		gatewayPostParams.put("x_delim_data", "TRUE");
		gatewayPostParams.put("x_delim_char", "|");
		gatewayPostParams.put("x_relay_response", "FALSE");
		gatewayPostParams.put("x_duplicate_window", getSettings().getValue(
				"authorizeNetDupeTxTimeout"));

		// transaction fields
		gatewayPostParams.put("x_method", "CC");
		gatewayPostParams.put("x_type", parameters.get("x_type"));
		gatewayPostParams.put("x_amount", parameters.get("x_amount"));
		gatewayPostParams.put("x_freight", parameters.get("x_freight"));
		gatewayPostParams.put("x_tax", parameters.get("x_tax"));
		if (getUser() != null && getUser().getOrder() != null) {
			if (parameters.get("x_amount") == null
					&& getUser().getOrder().getTotal() != null) {
				gatewayPostParams.put("x_amount", getUser().getOrder()
						.getTotal().toString());
			}
			if (parameters.get("x_freight") == null
					&& getUser().getOrder().getShipping() != null) {
				Double shipping = getUser().getOrder().getShipping();
				shipping = utils().roundForMoney(shipping);
				if (log.isDebugEnabled())
					log.debug("Sending shipping total: " + shipping);
				gatewayPostParams.put("x_freight", shipping.toString());
			}
			if (parameters.get("x_tax") == null
					&& getUser().getOrder().getTax() != null) {
				gatewayPostParams.put("x_tax", getUser().getOrder().getTax()
						.toString());
			}
		}
		gatewayPostParams.put("x_customer_ip", parameters.get("ipAddress"));
		gatewayPostParams.put("x_trans_id", parameters.get("x_trans_id"));
		gatewayPostParams.put("x_card_num", parameters.get("decryptedNumber"));
		if (parameters.get("creditCardExpirationMonth") != null
				&& parameters.get("creditCardExpirationYear") != null) {
			gatewayPostParams.put("x_exp_date", parameters.get(
					"creditCardExpirationMonth").toString()
					+ parameters.get("creditCardExpirationYear").toString());
		}
		if (parameters.get("creditCardSecurityCode") != null) {
			gatewayPostParams.put("x_card_code", parameters
					.get("creditCardSecurityCode"));
		}

		// invoice and customer reference fields
		gatewayPostParams.put("x_cust_id", parameters.get("customerID"));
		gatewayPostParams.put("x_invoice_num", parameters.get("invoiceNumber"));
		gatewayPostParams.put("x_description", parameters.get("description"));
		gatewayPostParams.put("x_po_num", parameters.get("poNumber"));

		// customer name and billing address fields
		if (getUser() != null && getUser().getOrder() != null) {
			if (getUser().getOrder().getFirstName() != null)
				gatewayPostParams.put("x_first_name", getUser().getOrder()
						.getFirstName());
			if (getUser().getOrder().getLastName() != null)
				gatewayPostParams.put("x_last_name", getUser().getOrder()
						.getLastName());
			if (getUser().getOrder().getOrganization() != null)
				gatewayPostParams.put("x_company", getUser().getOrder()
						.getOrganization());
			if (getUser().getOrder().getAddress1() != null)
				gatewayPostParams.put("x_address", getUser().getOrder()
						.getAddress1());
			if (getUser().getOrder().getAddress2() != null)
				gatewayPostParams.put("x_address", gatewayPostParams.get(
						"x_address").toString()
						+ " " + getUser().getOrder().getAddress2());
			if (getUser().getOrder().getCity() != null)
				gatewayPostParams.put("x_city", getUser().getOrder().getCity());
			if (getUser().getOrder().getState() != null)
				gatewayPostParams.put("x_state", getUser().getOrder()
						.getState());
			if (getUser().getOrder().getPostalCode() != null)
				gatewayPostParams.put("x_zip", getUser().getOrder()
						.getPostalCode());
			if (getUser().getOrder().getCountry() != null)
				gatewayPostParams.put("x_country", getUser().getOrder()
						.getCountry());
			if (getUser().getOrder().getPhone1() != null)
				gatewayPostParams.put("x_phone", getUser().getOrder()
						.getPhone1());
			if (getUser().getOrder().getPhone2() != null)
				gatewayPostParams
						.put("x_fax", getUser().getOrder().getPhone2());
			if (getUser().getOrder().getEmail1() != null)
				gatewayPostParams.put("x_email", getUser().getOrder()
						.getEmail1());
		}

		// customer shipping fields
		if (getUser() != null && getUser().getOrder() != null
				&& getUser().getOrder().firstDelivery() != null) {
			if (getUser().getOrder().firstDelivery().getFirstName() != null)
				gatewayPostParams.put("x_ship_to_first_name", getUser()
						.getOrder().firstDelivery().getFirstName());
			if (getUser().getOrder().firstDelivery().getLastName() != null)
				gatewayPostParams.put("x_ship_to_last_name", getUser()
						.getOrder().firstDelivery().getLastName());
			if (getUser().getOrder().firstDelivery().getOrganization() != null)
				gatewayPostParams.put("x_ship_to_company", getUser().getOrder()
						.firstDelivery().getOrganization());
			if (getUser().getOrder().firstDelivery().getAddress1() != null)
				gatewayPostParams.put("x_ship_to_address", getUser().getOrder()
						.firstDelivery().getAddress1());
			if (getUser().getOrder().firstDelivery().getAddress2() != null)
				gatewayPostParams.put("x_ship_to_address", gatewayPostParams
						.get("x_ship_to_address").toString()
						+ " "
						+ getUser().getOrder().firstDelivery().getAddress2());
			if (getUser().getOrder().firstDelivery().getCity() != null)
				gatewayPostParams.put("x_ship_to_city", getUser().getOrder()
						.firstDelivery().getCity());
			if (getUser().getOrder().firstDelivery().getState() != null)
				gatewayPostParams.put("x_ship_to_state", getUser().getOrder()
						.firstDelivery().getState());
			if (getUser().getOrder().firstDelivery().getPostalCode() != null)
				gatewayPostParams.put("x_ship_to_zip", getUser().getOrder()
						.firstDelivery().getPostalCode());
			if (getUser().getOrder().firstDelivery().getCountry() != null)
				gatewayPostParams.put("x_ship_to_country", getUser().getOrder()
						.firstDelivery().getCountry());
		}
		return gatewayPostParams;
	}

	public Map process(Map parameters) throws Exception {
		log.info("Processing payment.");

		// Transaction type from settings
		String authorizeNetTxType = (String) getSettings().getValue(
				"authorizeNetTxType");
		parameters.put("x_type", authorizeNetTxType);

		return processTransaction(parameters);
	}

	public Map processAuthOnly(Map parameters) throws Exception {
		log.info("Processing auth capture.");

		// transaction fields
		parameters.put("x_type", "AUTH_ONLY");
		return processTransaction(parameters);
	}

	public Map processAuthCapture(Map parameters) throws Exception {
		log.info("Processing auth capture.");

		// transaction fields
		parameters.put("x_type", "AUTH_CAPTURE");
		return processTransaction(parameters);
	}

	public Map processPriorAuthCapture(Map parameters) throws Exception {
		log.info("Processing prior auth capture.");

		// transaction fields
		parameters.put("x_type", "PRIOR_AUTH_CAPTURE");
		return processTransaction(parameters);
	}

	public Map processVoid(Map parameters) throws Exception {
		log.info("Processing void.");

		// transaction fields
		parameters.put("x_type", "VOID");
		return processTransaction(parameters);
	}

	protected Map processTransaction(Map parameters) throws Exception {
		log.info("Processing transaction.");

		// Set up common parameters for the request
		initializePostParams(parameters);

		// now send the request data to gateway
		String postData = prepareRequestData(gatewayPostParams, true);
		String urlString = (String) getSettings().getValue("authorizeNetURL");
		String responseData = sendHTTPPost(urlString, postData, false);

		// process the response
		return processResponse(responseData, parameters);
	}

	protected Map processResponse(String responseData, Map parameters)
			throws Exception {

		// parse response
		String[] responseArray = responseData.split("\\|");
		List response = java.util.Arrays.asList(responseArray);

		if (log.isDebugEnabled())
			printResponseArray(response);

		// process response, based on the Response Code in the first element
		// no response code indicates something like an HTTP 503 error code,
		// i.e., data that is exceptional. In this case, just return that string
		// on the result Map's resultMessage key.
		int responseCode = -1;
		try {
			responseCode = Integer.parseInt((String) response.get(0));
			log.info("Response Code: " + responseCode);
		} catch (Exception e) {
			log.error("No Response Code found.", e);
			return processErrorResponse(parameters, responseData);
		}
		switch (responseCode) {
		case RESPONSE_APPROVED:
			log.info("Payment Approved");
			return processSuccessResponse(parameters, response);
		case RESPONSE_DECLINED:
			log.info("Payment Declined");
			return processDeclinedResponse(parameters, response);
		case RESPONSE_ERROR:
		default:
			log.info("Payment Error");
			return processErrorResponse(parameters, response);
		}

	}

	protected Map processSuccessResponse(Map parameters, List response)
			throws Exception {
		Map result = new HashMap();
		result.put("resultCode", "0");
		result.put("resultMessage", response
				.get(AuthorizeNetAIMConstants.RESPONSE_REASON_TEXT));
		Payment payment = (Payment) getBusinessObjectFactory().createObject(
				"paymentImplementer");
		BeanUtils.copyProperties(payment, parameters);
		payment.setStatus("Received");
		try {
			String amountS = (String) response
					.get(AuthorizeNetAIMConstants.RESPONSE_AMOUNT);
			Double amountD = new Double(amountS);
			payment.setAmount(amountD);
		} catch (Exception e) {
			if (log.isWarnEnabled()) {
				log.warn("Error parsing payment amount: ", e);
			}
		}
		payment.setCreated(formatDateTime(new java.util.Date()));
		payment.setProcessorClassName(this.getClass().getName());
		String creditCardNumberDisplay = null;
		String decryptedNumber = payment.getDecryptedNumber();
		if (decryptedNumber != null) {
			creditCardNumberDisplay = "XXXXXXXXXXXX"
					+ decryptedNumber.substring(decryptedNumber.length() - 4,
							decryptedNumber.length());
			payment.setCreditCardNumberDisplay(creditCardNumberDisplay);
		}
		payment
				.setValueFieldLabels("TRANSACTION_TYPE,AMOUNT,CARD_CODE,RESPONSE_CODE,AVS_CODE,AUTHCODE,TRANSACTION_ID,REASON_CODE,REASON_TEXT");
		payment.setValue1((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_TRANSACTION_TYPE));
		payment.setValue2((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_AMOUNT));
		if (response.size() > AuthorizeNetAIMConstants.RESPONSE_CVV2_CODE) {
			payment.setValue3((String) response
					.get(AuthorizeNetAIMConstants.RESPONSE_CVV2_CODE));
		}
		payment.setValue4((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_CODE));
		payment.setValue5((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_AVS_CODE));
		payment.setValue6((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_AUTH_CODE));
		payment.setValue7((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_TRANSACTION_ID));
		payment.setValue8((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_REASON_CODE));
		payment.setValue9((String) response
				.get(AuthorizeNetAIMConstants.RESPONSE_REASON_TEXT));
		payment.setOrder(getUser().getOrder());
		if (getUser().getOrder().getPayments() == null)
			getUser().getOrder().setPayments(new LinkedHashSet());
		getUser().getOrder().getPayments().add(payment);
		PaymentDAO paymentDAO = (PaymentDAO) getDaoFactory().createDAO(
				"paymentDAOImplementer");
		paymentDAO.setPayment(payment);
		paymentDAO.insertPayment(false);
		result.put("newPayment", payment);
		return result;
	}

	protected Map processDeclinedResponse(Map parameters, List response)
			throws Exception {
		Map result = new HashMap();
		result.put("resultCode", response
				.get(AuthorizeNetAIMConstants.RESPONSE_REASON_CODE));
		result.put("resultMessage", response
				.get(AuthorizeNetAIMConstants.RESPONSE_REASON_TEXT));
		return result;
	}

	protected Map processErrorResponse(Map parameters, List response)
			throws Exception {
		Map result = new HashMap();
		result.put("resultCode", response
				.get(AuthorizeNetAIMConstants.RESPONSE_REASON_CODE));
		result.put("resultMessage", response
				.get(AuthorizeNetAIMConstants.RESPONSE_REASON_TEXT));
		return result;
	}

	protected Map processErrorResponse(Map parameters, String response)
			throws Exception {
		Map result = new HashMap();
		result.put("resultCode", "10000");
		return result;
	}

	protected void printResponseArray(List l) {
		log.debug("Printing response (" + l.size() + " elements)");
		for (int i = 0; i < l.size(); i++)
			log.debug(i + " " + l.get(i));
	}
}
