package com.softslate.commerce.businessobjects.payment.authorizeNetCIM;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import AuthNet.Rebill.CustomerAddressType;
import AuthNet.Rebill.CustomerPaymentProfileMaskedType;

import com.softslate.commerce.businessobjects.core.BusinessProcessor;
import com.softslate.commerce.businessobjects.customer.Customer;
import com.softslate.commerce.businessobjects.order.Order;
import com.softslate.commerce.businessobjects.payment.Payment;
import com.softslate.commerce.customer.order.BillingAddressForm;

/**
 * Interface that handles Authorize.net CIM profiles and payments
 * 
 * @author Jason McWilliams
 */

public interface AuthorizeNetCIMProcessor extends BusinessProcessor {

	public static final long CUSTOMER_IGNORE = -1;
	public static final long CUSTOMER_ANON = -2;
	public static final long PAYMENT_IGNORE = -1;

	abstract AuthorizeNetCIMResult deleteCustomerProfile(long profileID);

	/**
	 * Makes a call to pull the profile information for a user. For now, we
	 * populate the return object with a minimum amount of information. We may
	 * populate more information as needed.
	 * 
	 * @param profileID
	 */
	abstract AuthorizeNetCIMResultGetProfile getCustomerProfile(long profileID);

	/**
	 * Creates a customer profile. This will use information from a Customer or
	 * Payment bean to populate data. The beans do not have to be attached, as
	 * no information is changed. Caller should check the result object and
	 * update things if needed
	 * 
	 * @param customer
	 *            Information about the customer to make a profile for.
	 * @param orders
	 *            An optional collection of orders for the customer. Goes
	 *            through each order and tries to create payment profiles for
	 *            each payment. A list of paymentProfileIDs that were created
	 *            are returned in the result, but there is no real way to know
	 *            which payment they match. If it is important to mark a payment
	 *            record as recorded in CIM, first create a customer and then
	 *            create payment methods individually.
	 * @param liveOrder
	 *            Optional live order that can be used for address info if
	 *            needed.
	 * @throws Exception
	 */
	abstract AuthorizeNetCIMResultCreateProfile createCustomerProfile(
			Customer customer, Collection<Order> orders, Order liveOrder)
			throws Exception;

	/**
	 * Creates a customer profile. This will use information from an Order bean
	 * to populate the customer info. The bean does not have to be attached, as
	 * no information is changed. Caller should check the result object and
	 * update things if needed
	 * 
	 * @param order
	 * @throws Exception
	 */
	abstract AuthorizeNetCIMResultCreateProfile createCustomerProfile(
			Order order) throws Exception;

	/**
	 * Creates a payment profile on CIM for the incoming user using the payment
	 * bean for payment information and order bean for billing address
	 * information. We using the order information instead of the customer with
	 * the idea that the order is the latest by the user and will have the
	 * latest information.
	 * 
	 * @param cimProfileID
	 * @param payment
	 * @param order
	 * @throws Exception
	 */
	abstract AuthorizeNetCIMResult createCustomerPaymentProfile(
			long cimProfileID, Payment payment, Order order) throws Exception;

	/**
	 * Creates a payment profile on CIM for the incoming user using the map of
	 * information provided.
	 * 
	 * @param cimProfileID
	 * @param paymentParams
	 * @param billing
	 * @throws Exception
	 */
	abstract AuthorizeNetCIMResult createCustomerPaymentProfile(long cimProfileID, Map paymentParams, boolean billing)
			throws Exception;

	/**
	 * Tries to update a current payment profile with new information.
	 * 
	 * @param cimProfileID
	 *            Customer profile ID
	 * @param paymentProfileID
	 *            Payment profile ID
	 * @param payment
	 *            Payment bean containing the newest card data
	 * @param order
	 *            Order associated with the card to get newest billing info
	 * @throws Exception
	 */
	abstract AuthorizeNetCIMResult updateCustomerPaymentProfile(
			long cimProfileID, long paymentProfileID, Payment payment,
			Order order) throws Exception;

	/**
	 * Deletes a payment profile ID given a customer and payment profile
	 * belonging to that customer
	 * 
	 * @param profileID
	 * @param paymentProfileID
	 */
	abstract AuthorizeNetCIMResult deleteCustomerPaymentProfile(long profileID,
			long paymentProfileID);

	/**
	 * Loads list of objects that contain customerIDs that do not have a payment
	 * profile that are between this many days old
	 * 
	 * @param createdStart
	 *            a string date
	 * @param createdEnd
	 *            a string date
	 * @return
	 * @throws Exception
	 */
	abstract List<Object[]> loadCustomersWithNoCIMProfileByDate(
			String createdStart, String createdEnd) throws Exception;
			
	/**
	 * Loads collection of cutomers that do not have a CIM profile and were
	 * created after the date sent in.
	 * 
	 * @param cutoffDate
	 *            a string date
	 * @throws Exception
	 */
	abstract Collection<Customer> loadCustomersWithNoCIMProfileByDate(
			String cutoffDate) throws Exception;

	/**
	 * Loads collection of payment ids that do not yet have a paymentProfileID
	 * and have a credit card number. This would likely be used to send over
	 * payment information in a batch job. The profiles must have been created
	 * between these dates.
	 * 
	 * @@param createdStart a string date
	 * @param createdEnd
	 *            a string date
	 * @return
	 * @throws Exception
	 */
	abstract List<Object[]> loadPaymentsWithNoProfileByDate(
			String createdStart, String createdEnd) throws Exception;
			
	/**
	 * Loads collection of payments that do not yet have a paymentProfileID and
	 * have a credit card number. This would likely be used to send over payment
	 * information in a batch job.
	 * 
	 * @param cutoffDate
	 *            a string date to look backwards in time
	 * @throws Exception
	 */
	abstract Collection<Payment> loadPaymentsWithNoCIMProfileByDate(
			String cutoffDate) throws Exception;

	/**
	 * Return a map that has xmlResultCode, xmlMessageCode, xmlMessageText,
	 * token The result code should not have an error and the token should be
	 * populated in order to be a valid response.
	 * 
	 * @param cimProfileID
	 * @return
	 * @throws Exception
	 */
	abstract Map<String, String> getHostedProfilePage(Long cimProfileID)
			throws Exception;
			
	/**
	 * Return a map that has xmlResultCode, xmlMessageCode, xmlMessageText,
	 * token The result code should not have an error and the token should be
	 * populated in order to be a valid response.
	 * 
	 * @param cimProfileID
	 * @param fromWhere
	 *            Where are we calling the page from, in case of redirect.
	 * @throws Exception
	 */
	abstract Map<String, String> getHostedProfilePage(Long cimProfileID,
			String fromWhere) throws Exception;

	/**
	 * Checks to see if a payment method chosen by the user is actually valid.
	 * Uses getUser().getOrder() and getUser().getCustomer() to check
	 * everything.
	 * 
	 * @param parameters
	 * @throws Exception
	 */
	abstract Map<String, String> validatePaymentMethod(
			Map<String, String> parameters) throws Exception;

	/**
	 * Set anything up as needed
	 */
	abstract void initialize();

	/**
	 * Copy a current cim profile into a billing address form
	 * 
	 * @param payment
	 * @param form
	 */
	abstract void copyPaymentPaymentProfileToAddressForm(AuthorizeNetCIMPaymentProfileMin payment,
			BillingAddressForm form);

	/**
	 * Updates a customer payment profile from an existing masked type. The
	 * caller should get the customer profiles, update the masked type they want
	 * to update, and then send it in here for updating.
	 * 
	 * @param cimProfileID
	 * @param paymentProfile
	 * @throws Exception
	 */
	abstract AuthorizeNetCIMResult updateCustomerPaymentProfile(long cimProfileID,
			CustomerPaymentProfileMaskedType paymentProfile) throws Exception;

	/**
	 * Builds CustomerAddressType from a map of address fields, normally this
	 * would be from a softslate address form.
	 * 
	 * @param parameters
	 * @param billing
	 *            Should we look at params starting only with billing*
	 * @throws Exception
	 */
	abstract CustomerAddressType buildCustomerAddressType(Map parameters, boolean billing) throws Exception;

	/**
	 * Find a matching payment profile for the given customer and profile id.
	 * This looks through the customers information on auth.net to try to match
	 * the profile the user wants to edit, delete, etc. Will return null if not
	 * found and log why.
	 * 
	 * @param customer
	 * @param order
	 * @param incomingPaymentProfile
	 * @throws Exception
	 */
	abstract AuthorizeNetCIMPaymentProfileMin findMatchingPaymentProfile(Customer customer, Order order,
			String incomingPaymentProfile) throws Exception;

	/**
	 * Find collection of payment profiles for a customer/user
	 * 
	 * @param customer
	 * @param order
	 * @throws Exception
	 */
	abstract Collection<AuthorizeNetCIMPaymentProfileMin> findPaymentProfiles(Customer customer, Order order)
			throws Exception;

	/**
	 * Checks the customer and then the order for a valid existing CIM profile
	 * id.
	 * 
	 * @param customer
	 * @param order
	 */
	abstract long findValidCimProfileID(Customer customer, Order order);

	abstract AuthorizeNetCIMResult processPriorAuthCapture(Order order, Map parameters) throws Exception;

	abstract Map<String, String> logProfileErrors(String internalErrorMessage, List<Map<String, String>> messages,
			Customer customer, Order order, String command);

	abstract Map processSuccessResponse(Map parameters, AuthorizeNetCIMResult authorizeNetCIMResult,
			boolean priorAuthCapture) throws Exception;
}
