package com.softslate.commerce.businessobjects.product;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.BaseBusinessProcessor;
import com.softslate.commerce.businessobjects.order.OrderItem;
import com.softslate.commerce.daos.product.DiscountDAO;
import com.softslate.commerce.daos.product.DiscountGatewayDAO;
import com.softslate.commerce.daos.product.SKUDAO;

/**
 * Processes business logic having to do with discounts. Created in the Struts
 * layer in response to requests for viewing, editing, adding, and deleting
 * discounts.
 * <p>
 * <code>BasicDiscountProcessor</code> is the default implementation of the
 * <code>DiscountProcessor</code> interface for the application.
 * 
 * @author David Tobey
 */
public class BasicDiscountProcessor extends BaseBusinessProcessor implements
		DiscountProcessor {

	static Log log = LogFactory.getLog(BasicDiscountProcessor.class);

	public Collection loadStoreDiscounts() throws Exception {
		if (log.isInfoEnabled())
			log.info("Loading store discounts");
		DiscountGatewayDAO gatewayDAO =  getDaoFactory()
				.createDAO(DiscountGatewayDAO.class);
		return gatewayDAO.loadStoreDiscounts();
	}

	public Map loadDiscountsAndCount(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName()
					+ ": Loading discounts and count");
		DiscountGatewayDAO gatewayDAO =  getDaoFactory()
				.createDAO(DiscountGatewayDAO.class);
		return gatewayDAO.loadDiscountsAndCount(parameters);
	}

	public Map addDiscount(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("code");
			log.info(getUser().getAdministrator().getUserName()
					+ ": Adding discount: " + code);
		}
		Map results = new HashMap();

		// Check that another discount with the same code doesn't exist
		Discount discount = (Discount) getBusinessObjectFactory().createObject(
				"discountImplementer");
		BeanUtils.copyProperties(discount, parameters);
		results = validateDiscount(discount);
		if (results != null && results.get("resultCode") != null
				&& !results.get("resultCode").equals("0")) {
			return results;
		}

		// Insert the discount
		if (discount.getOrderItemID() != null
				&& discount.getOrderItemID().intValue() > 0) {
			OrderItem orderItem =  getBusinessObjectFactory()
					.createObject(OrderItem.class);
			orderItem.setOrderItemID(discount.getOrderItemID().intValue());
			discount.setOrderItem(orderItem);
		}
		DiscountDAO dao = (DiscountDAO) getDaoFactory().createDAO(
				"discountDAOImplementer");
		dao.setDiscount(discount);
		dao.insertDiscount(true);
		results.put("newRecord", dao.getDiscount());

		// Insert the association with the parent product, if present
		if (parameters.get("productID") != null
				&& !parameters.get("productID").equals("")) {

			// First see if there is a product sku set up yet
			SKU sku = (SKU) getBusinessObjectFactory().createObject(
					"skuImplementer");
			BeanUtils.copyProperties(sku, parameters);
			SKUDAO skuDao = (SKUDAO) getDaoFactory().createDAO(
					"skuDAOImplementer");
			skuDao.setSKU(sku);
			skuDao.loadProductSKU();
			if (skuDao.getSKU() != null) {
				skuDao.getDaoFactory().getSession().evict(skuDao.getSKU());
			}
			sku = skuDao.getSKU();

			// If not, attempt to insert the product sku now.
			if (sku == null) {

				ProductProcessor processor =  getBusinessObjectFactory()
						.createObject(ProductProcessor.class);
				Product product = processor.loadProductFromID(parameters);

				SKUProcessor skuProcessor =  getBusinessObjectFactory()
						.createObject(SKUProcessor.class);
				Map skuParameters = new HashMap();
				skuParameters.put("productID", parameters.get("productID"));
				skuParameters.put("skuNumber", product.getCode());
				skuParameters.put("skuName", product.getName());
				skuParameters.put("trackInventory", "false");

				Map skuResults = skuProcessor.addSKU(skuParameters);
				sku = (SKU) skuResults.get("newRecord");
			}

			// Now attempt to insert an association between the discount and the
			// sku.
			if (sku != null) {
				SKUDiscount sd =  getBusinessObjectFactory()
						.createObject(SKUDiscount.class);
				DiscountGatewayDAO gatewayDAO =  getDaoFactory()
						.createDAO(DiscountGatewayDAO.class);
				sd.setSku(sku);
				sd.setDiscount(dao.getDiscount());
				gatewayDAO.insertSKUDiscount(sd, true);
			}
		}

		return results;
	}

	public Discount loadDiscountFromID(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName()
					+ ": Loading discount from id");
		Discount discount = (Discount) getBusinessObjectFactory().createObject(
				"discountImplementer");
		DiscountDAO dao = (DiscountDAO) getDaoFactory().createDAO(
				"discountDAOImplementer");
		BeanUtils.copyProperties(discount, parameters);
		dao.setDiscount(discount);
		dao.loadDiscountFromID();
		return dao.getDiscount();
	}

	public Map editDiscount(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("code");
			log.info(getUser().getAdministrator().getUserName()
					+ ": Editing discount: " + code);
		}
		Map results = new HashMap();

		// Check that another discount with the same code doesn't exist
		Discount discount = (Discount) getBusinessObjectFactory().createObject(
				"discountImplementer");
		BeanUtils.copyProperties(discount, parameters);
		results = validateDiscount(discount);
		if (results != null && results.get("resultCode") != null
				&& !results.get("resultCode").equals("0")) {
			return results;
		}

		// Load the discount
		discount = loadDiscountFromID(parameters);
		BeanUtils.copyProperties(discount, parameters);

		// Update the discount
		if (discount.getOrderItemID() != null
				&& discount.getOrderItemID().intValue() > 0) {
			OrderItem orderItem =  getBusinessObjectFactory()
					.createObject(OrderItem.class);
			orderItem.setOrderItemID(discount.getOrderItemID().intValue());
			discount.setOrderItem(orderItem);
		}
		DiscountDAO dao = (DiscountDAO) getDaoFactory().createDAO(
				"discountDAOImplementer");
		dao.setDiscount(discount);
		dao.updateDiscount(true);
		results.put("updatedRecord", dao.getDiscount());
		return results;
	}

	public Map editDeleteDiscounts(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName() + ": Editing/deleting multiple discounts");
		DiscountGatewayDAO dao =  getDaoFactory().createDAO(DiscountGatewayDAO.class);
		Map results = dao.editDeleteDiscounts(parameters, getBusinessObjectFactory());
		return results;
	}

	public void deleteDiscount(Map parameters) throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null) {
			String code = (String) parameters.get("code");
			log.info(getUser().getAdministrator().getUserName()
					+ ": Deleting discount: " + code);
		}
		Discount discount = (Discount) getBusinessObjectFactory().createObject(
				"discountImplementer");
		BeanUtils.copyProperties(discount, parameters);
		DiscountDAO dao = (DiscountDAO) getDaoFactory().createDAO(
				"discountDAOImplementer");
		dao.setDiscount(discount);
		dao.deleteDiscount(true);
		return;
	}

	public Map validateDiscount(Discount discount) throws Exception {
		Map results = new HashMap();
		DiscountDAO dao = (DiscountDAO) getDaoFactory().createDAO(
				"discountDAOImplementer");
		dao.setDiscount(discount);
		dao.loadDiscountFromCode();
		if (dao.getDiscount() != null) {
			dao.getDaoFactory().getSession().evict(dao.getDiscount());
		}
		boolean insertingButExists = discount.getDiscountID() == 0
				&& dao.getDiscount() != null
				&& dao.getDiscount().getDiscountID() > 0;
		boolean updatingButExists = discount.getDiscountID() > 0
				&& dao.getDiscount() != null
				&& discount.getDiscountID() != dao.getDiscount()
						.getDiscountID();
		if (insertingButExists || updatingButExists) {
			results.put("resultCode", "DuplicateRecord");
			results.put("duplicateField", "code");
			results.put("existingRecord", dao.getDiscount());
		}
		return results;
	}

	public Collection loadAllCoupons() throws Exception {
		if (log.isInfoEnabled() && getUser().getAdministrator() != null)
			log.info(getUser().getAdministrator().getUserName()
					+ ": Loading all coupons");
		DiscountGatewayDAO gatewayDAO =  getDaoFactory()
				.createDAO(DiscountGatewayDAO.class);
		return gatewayDAO.loadAllCoupons();
	}
}
