package com.softslate.commerce.daos.core;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.HibernateException;
import org.hibernate.MappingException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.boot.registry.StandardServiceRegistryBuilder;
import org.hibernate.cfg.Configuration;
import org.hibernate.exception.ConstraintViolationException;

import com.google.inject.Injector;
import com.softslate.commerce.businessobjects.core.Settings;

/**
 * Factory class used to create and initialize instances of Hibernate data
 * access objects.
 * <p>
 * An instance of <code>HDAOFactory</code> is created in the Struts layer and
 * placed in the application scope. (See
 * {@link com.softslate.commerce.customer.core.BaseRequestProcessor}.)
 * <p>
 * Immediately after instantiating the requested DAO, this factory passes its
 * own properties to it so that the newly created object can use them to
 * communicate with the rest of the application.
 * 
 * @author David Tobey
 */
public class HibernateDAOFactory extends BaseDAOFactory implements DAOFactory {

	static Log log = LogFactory.getLog(HibernateDAOFactory.class);

	private static final ThreadLocal threadSession = new ThreadLocal();

	private static final ThreadLocal threadTransaction = new ThreadLocal();

	private Settings settings = null;

	public Settings getSettings() {
		return (this.settings);
	}

	public void setSettings(Settings settings) {
		this.settings = settings;
	}

	private Properties appSettings = null;

	public Properties getAppSettings() {
		return appSettings;
	}

	public void setAppSettings(Properties appSettings) {
		this.appSettings = appSettings;
	}

	private Properties appComponents = null;

	public Properties getAppComponents() {
		return appComponents;
	}

	public void setAppComponents(Properties appComponents) {
		this.appComponents = appComponents;
	}

	private Properties appComponentsDefaults = null;

	public Properties getAppComponentsDefaults() {
		return appComponentsDefaults;
	}

	public void setAppComponentsDefaults(Properties appComponentsDefaults) {
		this.appComponentsDefaults = appComponentsDefaults;
	}

	private Injector injector = null;

	public Injector getInjector() {
		return injector;
	}

	public void setInjector(Injector injector) {
		this.injector = injector;
	}

	protected SessionFactory sessionFactory = null;

	public SessionFactory getSessionFactory() {
		return sessionFactory;
	}

	public void setSessionFactory(SessionFactory sessionFactory) {
		this.sessionFactory = sessionFactory;
	}

	private Map importExportSqlMap = null;

	public Map getImportExportSqlMap() {
		return importExportSqlMap;
	}

	public void setImportExportSqlMap(Map importExportSqlMap) {
		this.importExportSqlMap = importExportSqlMap;
	}

	public void initialize() throws Exception {
		// Initialize the Hibernate SessionFactory.
		initializeSessionFactory();
	}

	/**
	 * Creates a configured instance of Hibernate's <code>SessionFactory</code>
	 * in the application scope.
	 */
	public Configuration initializeSessionFactory() {
		if (log.isInfoEnabled())
			log.info("Initializing Hibernate SessionFactory.");

		// Set the database table prefix
		Configuration configuration = new Configuration();
		HibernateNamingStrategy hns = new HibernateNamingStrategy();
		hns.tablePrefix = getAppSettings().getProperty("databaseTablePrefix");
		if (log.isInfoEnabled())
			log.info("getAppSettings(): " + getAppSettings());
		configuration.setNamingStrategy(hns);

		// Add each component class's mapping file to the SessionFactory.
		// Start with the defaults so that we may use the extends attribute in a
		// subclass tag for a custom component.
		addComponentMappings(configuration, getAppComponentsDefaults());
		addComponentMappings(configuration, getAppComponents());

		// Add named queries
		configuration.addResource("resources/administrator/queries.hbm.xml");
		configuration
				.addResource("resources/administrator/queries-custom.hbm.xml");
		configuration.addResource("resources/core/queries.hbm.xml");
		configuration.addResource("resources/core/queries-custom.hbm.xml");
		configuration.addResource("resources/customer/queries.hbm.xml");
		configuration.addResource("resources/customer/queries-custom.hbm.xml");
		configuration.addResource("resources/email/queries.hbm.xml");
		try {
			configuration.addResource("resources/email/queries-custom.hbm.xml");
		} catch (org.hibernate.MappingNotFoundException mnfe) {
			log.warn(mnfe.toString() + ". If you're upgrading to 3.3.1 or later, add "
					+ "classes/resources/email/queries-custom.hbm.xml to avoid this warning.");
		}
		configuration.addResource("resources/order/queries.hbm.xml");
		configuration.addResource("resources/order/queries-custom.hbm.xml");
		configuration.addResource("resources/payment/queries.hbm.xml");
		configuration.addResource("resources/payment/queries-custom.hbm.xml");
		configuration.addResource("resources/product/queries.hbm.xml");
		configuration.addResource("resources/product/queries-custom.hbm.xml");
		configuration.addResource("resources/shipping/queries.hbm.xml");
		configuration.addResource("resources/shipping/queries-custom.hbm.xml");
		configuration.addResource("resources/tax/queries.hbm.xml");
		configuration.addResource("resources/tax/queries-custom.hbm.xml");

		// Set hibernate.listeners.envers.autoRegister to false, unless it's deliberately set to true (reverse the
		// regular behavior), so folks have to enable it deliberately.
		Properties configurationProps = getHibernateSettings();
		if (configurationProps.getProperty("hibernate.listeners.envers.autoRegister") == null) {
			configurationProps.setProperty("hibernate.listeners.envers.autoRegister", "false");
		}
		configuration.setProperties(getHibernateSettings());
		StandardServiceRegistryBuilder builder = new StandardServiceRegistryBuilder().applySettings(configuration
				.getProperties());
		SessionFactory sessionFactory = configuration.buildSessionFactory(builder.build());
		setSessionFactory(sessionFactory);

		return configuration;
	}

	public void addComponentMappings(Configuration configuration,
			Properties components) {
		Collection keySet = components.keySet();
		Iterator i = keySet.iterator();
		while (i.hasNext()) {
			String key = (String) i.next();
			String[] classNames = ((String) components.get(key)).split(",");
			for (int j = 0; j < classNames.length; j++) {
				String className = classNames[j].trim();
				if (!className.endsWith(".jsp") && className.length() > 0 && !key.endsWith("ProcessorImplementer")
						&& !key.endsWith("HelperImplementer") && !key.endsWith("DAOImplementer")
						&& !key.endsWith("TriggerImplementer") && !key.endsWith("EventImplementer")
						&& !key.endsWith("HandlerImplementer")) {
					try {
						Class classDefinition = Class.forName(className);
						configuration.addAnnotatedClass(classDefinition);
						configuration.addClass(classDefinition);
					} catch (ClassNotFoundException cnfe) {
						log.error("Class for component " + key + " not found: " + cnfe.toString(), cnfe);
					} catch (MappingException me) {
						log.debug("Hibernate mapping file for component " + key + " skipped: " + me.toString());
					}
				}
			}
		}
	}

	/**
	 * Creates and initializes a data access object for use by the application.
	 * Looks up the class to be instantiated by using the
	 * <code>appComponents</code> property.
	 * 
	 * @param type
	 *            A <code>String</code> associated with the class name of the
	 *            data access object to be created.
	 * @return An instantiated and initialized data access object, or
	 *         <code>null</code> if initialization failed.
	 * @throws Exception
	 */
	public Object createDAO(String type) throws Exception {
		String className = getAppComponents().getProperty(type);
		return createDAOFromClassName(className);
	}

	public <T> T createDAO(Class<T> type) {
		try {
			if (!type.isInterface()) {
				return (T) createDAOFromClassName(type.getName());
			}
			return (T) createDAO(findImplementer(type.getSimpleName()));
		} catch (RuntimeException e) {
			throw e;
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	public String findImplementer(String type) {
		if (type.startsWith("Custom") && Character.isUpperCase(type.charAt("Custom".length()))) {
			type = type.substring("Custom".length());
		}
		if (type.endsWith("Impl")) {
			type = type.substring(0, type.length() - "Impl".length());
		}
		if (type.startsWith("SKU")) {
			type = type.substring(0, "SKU".length()).toLowerCase() + type.substring("SKU".length());
		}
		return type.substring(0, 1).toLowerCase() + type.substring(1) + "Implementer";
	}

	/**
	 * Retrieves a <code>Map</code> of properties that will be transfered to
	 * each of the data access objects this factory creates.
	 * 
	 * @return A <code>Map</code> of properties that will be transfered to
	 *         each of the data access objects this factory creates
	 */
	public Map getTransferMap() {
		Map map = new HashMap();
		map.put("settings", getSettings());
		map.put("appSettings", getAppSettings());
		map.put("appComponents", getAppComponents());
		map.put("sqlMap", getSqlMap());
		map.put("importExportSqlMap", getImportExportSqlMap());
		map.put("dataSource", getDataSource());
		map.put("injector", getInjector());
		map.put("daoFactory", this);
		return map;
	}

	/**
	 * Starts an interaction with Hibernate.
	 * 
	 * @return Session
	 */
	public Session startSession() {
		log.debug("startSession()");
		Session s = getSession();
		beginTransaction();
		return s;
	}

	/**
	 * Retrieves the current Session local to the thread.
	 * 
	 * @return Session
	 */
	public Session getSession() {
		log.debug("getSession()");
		Session s = (Session) threadSession.get();
		if (s == null) {
			log.debug("Opening Session for this thread.");
			s = getSessionFactory().openSession();
			threadSession.set(s);
		}
		return s;
	}

	/**
	 * Closes the Session local to the thread.
	 */
	public void closeSession() {
		log.debug("closeSession()");
		Session s = (Session) threadSession.get();
		threadSession.set(null);
		if (s != null && s.isOpen()) {
			log.debug("Closing Session for this thread.");
			s.close();
		}
	}

	/**
	 * Start a new database transaction.
	 */
	public void beginTransaction() {
		log.debug("beginTransaction()");
		Transaction tx = (Transaction) threadTransaction.get();
		if (tx == null) {
			log.debug("Starting new transaction for this thread.");
			tx = getSession().beginTransaction();
			threadTransaction.set(tx);
		}
	}

	/**
	 * Commit the database transaction.
	 */
	public void commitTransaction() throws HibernateException {
		log.debug("commitTransaction()");
		Transaction tx = (Transaction) threadTransaction.get();
		try {
			if (tx != null && !tx.wasCommitted() && !tx.wasRolledBack()) {
				log.debug("Commit transaction of this thread.");
				tx.commit();
			}
			threadTransaction.set(null);
		} catch (ConstraintViolationException cve) {
			log.error("Constraint violation while committing: ", cve);
			rollbackTransaction();
			throw cve;
		} catch (HibernateException ex) {
			log.error("Exception committing: ", ex);
			rollbackTransaction();
			throw ex;
		} catch (org.hibernate.AssertionFailure af) {
			log.error("Exception committing: ", af);
			rollbackTransaction();
			throw af;
		}
	}
	
	/**
	 * Rollback the database transaction.
	 */
	public void rollbackTransactionKeepingSessionOpen() {
		log.debug("rollbackTransactionKeepingSessionOpen()");
		Transaction tx = (Transaction) threadTransaction.get();
		threadTransaction.set(null);
		if (tx != null && !tx.wasCommitted() && !tx.wasRolledBack()) {
			log.debug("Rollback transaction of this thread.");
			tx.rollback();
		}
	}

	/**
	 * Rollback the database transaction.
	 */
	public void rollbackTransaction() {
		log.debug("rollbackTransaction()");
		Transaction tx = (Transaction) threadTransaction.get();
		try {
			threadTransaction.set(null);
			if (tx != null && !tx.wasCommitted() && !tx.wasRolledBack()) {
				log.debug("Rollback transaction of this thread.");
				tx.rollback();
			}
		} finally {
			log.debug("closing session");
			closeSession();
		}
	}

	public String getDatabaseType() {
		if (getHibernateSettings() != null
				&& getHibernateSettings().getProperty("hibernate.dialect") != null) {
			String dialect = getHibernateSettings().getProperty(
					"hibernate.dialect");
			if (dialect.contains("MySQL")) {
				return "MySQL";
			}
			if (dialect.contains("PostgreSQL")) {
				return "Postgres";
			}
			if (dialect.contains("Oracle")) {
				return "Oracle";
			}
			if (dialect.contains("SQLServer")) {
				return "MSSQL";
			}
			dialect = dialect.replaceFirst("org.hibernate.dialect.", "");
			dialect = dialect.replaceFirst("Dialect", "");
			return dialect;
		}
		return null;
	}

	/**
	 * Evicts the Hibernate 2nd level cache, plus any cached queries. Evicts
	 * each of the cached regions defined in the default version of ehcache.xml,
	 * using <code>getSessionFactory().evict(String region)</code>, and then
	 * runs <code>getSessionFactory().evictQueries();</code> to evicted cached
	 * queries.
	 */
	public void evictCache() {
		getSessionFactory().getCache().evictEntityRegion(com.softslate.commerce.businessobjects.product.Category.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.CategorySetting.class);
		getSessionFactory().getCache().evictEntityRegion(com.softslate.commerce.businessobjects.product.Product.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.ProductSetting.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.ProductCategory.class);
		getSessionFactory().getCache()
				.evictEntityRegion(com.softslate.commerce.businessobjects.product.Attribute.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.ProductAttribute.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.RelatedProduct.class);
		getSessionFactory().getCache().evictEntityRegion(com.softslate.commerce.businessobjects.product.Option.class);
		getSessionFactory().getCache().evictEntityRegion(com.softslate.commerce.businessobjects.product.SKU.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.SKUDiscount.class);
		getSessionFactory().getCache().evictEntityRegion(com.softslate.commerce.businessobjects.product.Discount.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.DiscountRange.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.product.Manufacturer.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.shipping.ShippingRule.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.shipping.ShippingRuleRange.class);
		getSessionFactory().getCache().evictEntityRegion(
				com.softslate.commerce.businessobjects.shipping.SKUShippingRule.class);
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Category.subcategoryCollection");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Category.products");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Category.productCategories");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Category.categorySettings");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Product.productCategories");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Product.productAttributes");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Product.productSettings");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Product.skus");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Product.relatedProducts");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Attribute.productAttributes");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Attribute.attributeOptions");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.SKU.skuDiscounts");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Discount.skuDiscounts");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Discount.discountRanges");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.Manufacturer.products");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.product.SKU.skuShippingRules");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.shipping.ShippingRule.skuShippingRules");
		getSessionFactory().getCache().evictCollectionRegion(
				"com.softslate.commerce.businessobjects.shipping.ShippingRule.shippingRuleRanges");
		getSessionFactory().getCache().evictQueryRegions();
	}

}
