package com.softslate.commerce.daos.customer;

import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.Hibernate;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;

import com.softslate.commerce.businessobjects.customer.Customer;
import com.softslate.commerce.businessobjects.customer.CustomerWishList;
import com.softslate.commerce.daos.core.BaseDAO;
import com.softslate.commerce.daos.core.DataAccessException;

/**
 * Generic DAO to manage the <code>sscCustomer</code> table of the database.
 * 
 * @author David Tobey
 * 
 */
public class CustomerDAOHibernate extends BaseDAO implements CustomerDAO {

	static Log log = LogFactory.getLog(CustomerDAOHibernate.class);

	private Customer customer = null;

	public Customer getCustomer() {
		return (this.customer);
	}

	public void setCustomer(Customer customer) {
		this.customer = customer;
	}

	public void loadCustomerFromID() throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			Customer record = (Customer) session.get(Customer.class,
					new Integer(getCustomer().getCustomerID()));
			if (record != null) {
				if (record.getPassword() != null) {
					record.setDecryptedPassword(getDAOUtils().getDecrypted(
							record.getPassword(),
							(String) getSettings().getValue(
									"customerEncryptionType"),
							getAppSettings().getProperty("twoWayKeyFile")));
				}
				PropertyUtils.copyProperties(getCustomer(), record);
			}
			setCustomer(record);
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
		return;
	}

	public String loadPasswordByID() throws DataAccessException {
		Session session = getDaoFactory().startSession();
		try {
			Query q = getNamedQuery(session, "customer.selectPasswordFromID");
			q.setInteger("customerID", getCustomer().getCustomerID());
			String password = (String) q.uniqueResult();
			return password;
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
	}

	public int getUserNameCount() throws Exception {
		return getUserNameCount(false);
	}

	public int getUserNameCount(boolean commit) throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			Query q = getNamedQuery(session, "customer.selectUserNameCount");
			q.setString("userName", getCustomer().getUserName());
			q.setInteger("customerID", getCustomer().getCustomerID());
			Long count = (Long) q.uniqueResult();
			if (commit) {
				getDaoFactory().commitTransaction();
			}
			return count.intValue();
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
	}

	public void insertCustomer() throws DataAccessException {
		insertCustomer(false);
	}

	public void insertCustomer(boolean commit) throws DataAccessException {
		if (getCustomer().getDecryptedPassword() != null) {
			getCustomer().setPassword(
					getDAOUtils().getEncrypted(
							getCustomer().getDecryptedPassword(),
							(String) getSettings().getValue(
									"customerEncryptionType"),
							getAppSettings().getProperty("twoWayKeyFile"),
							(String) getSettings().getValue("oneWayEncryptionAlgorithm")));
		}
		if (getCustomer().getPasswordToken() != null && getCustomer().getPasswordToken().equals("")) {
			getCustomer().setPasswordToken(null);
		}
		Session session = getDaoFactory().startSession();
		try {
			session.save(getCustomer());
			if (commit) {
				getDaoFactory().commitTransaction();
			}
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
		return;
	}

	public void updateCustomer() throws DataAccessException {
		updateCustomer(false, false);
	}

	public void updateCustomer(boolean commit) throws DataAccessException {
		updateCustomer(commit, false);
	}

	public void updateCustomer(boolean commit, boolean forcePassword)
			throws DataAccessException {
		// Do not re-encrypt the password if the encryption type is one way.
		String encType = (String) getSettings().getValue(
				"customerEncryptionType");
		if (forcePassword
				|| (encType != null
						&& !encType.equals("oneWay")
						&& getCustomer().getDecryptedPassword() != null
						&& getCustomer().getDecryptedPassword().length() > 0
						&& !getCustomer().getDecryptedPassword().equals(
								"[noPermission]")
						&& !getCustomer().getDecryptedPassword().equals(
								"[oneWayEncryption]") && !getCustomer()
						.getDecryptedPassword().equals(
								"[keyPairNoStoreEncryption]"))) {
			getCustomer().setPassword(
					getDAOUtils().getEncrypted(
							getCustomer().getDecryptedPassword(), encType,
							getAppSettings().getProperty("twoWayKeyFile"),
							(String) getSettings().getValue("oneWayEncryptionAlgorithm")));
		} else {
			// Use the password that's already in the database.
			try {
				String password = loadPasswordByID();
				getCustomer().setPassword(password);
			} catch (Exception e) {
			}
		}
		if (getCustomer().getPasswordToken() != null && getCustomer().getPasswordToken().equals("")) {
			getCustomer().setPasswordToken(null);
		}
		Session session = getDaoFactory().startSession();
		try {
			session.update(getCustomer());
			if (commit) {
				getDaoFactory().commitTransaction();
			}
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
		return;
	}

	/**
	 * Assuming the userName property has been set, queries the database for a
	 * customer record that matches.
	 */
	public void loadFromUserName() throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			Query q = getNamedQuery(session, "customer.selectFromUserName");
			q.setString("userName", getCustomer().getUserName());
			setCustomer((Customer) q.uniqueResult());
			if (getCustomer() == null) {
				return;
			}
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
		if (getCustomer().getPassword() != null) {
			getCustomer().setDecryptedPassword(
					getDAOUtils().getDecrypted(
							getCustomer().getPassword(),
							(String) getSettings().getValue(
									"customerEncryptionType"),
							getAppSettings().getProperty("twoWayKeyFile")));
		}
		return;
	}

	/**
	 * Assuming the userName and password properties have been set, queries the
	 * database for a customer record that matches.
	 */
	public void loadFromUserNameAndPassword() throws Exception {
		if (getCustomer().getDecryptedPassword() != null) {
			getCustomer().setPassword(
					getDAOUtils().getEncrypted(
							getCustomer().getDecryptedPassword(),
							(String) getSettings().getValue(
									"customerEncryptionType"),
							getAppSettings().getProperty("twoWayKeyFile"),
							(String) getSettings().getValue("oneWayEncryptionAlgorithm")));
		}
		Session session = getDaoFactory().startSession();
		String decryptedPassword = getCustomer().getDecryptedPassword();
		try {
			Query q = getNamedQuery(session,
					"customer.selectFromUserNameAndPassword");
			q.setString("userName", getCustomer().getUserName());
			q.setString("password", getCustomer().getPassword());
			setCustomer((Customer) q.uniqueResult());
			if (getCustomer() == null) {
				return;
			}
			getCustomer().setDecryptedPassword(decryptedPassword);
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
		return;
	}

	/**
	 * Assuming the autoLoginToken has been set, queries the database for a
	 * customer record that matches.
	 */
	public void loadFromAutoLoginToken() throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			Query q = getNamedQuery(session,
					"customer.selectFromAutoLoginToken");
			q.setString("autoLoginToken", getCustomer().getAutoLoginToken());
			setCustomer((Customer) q.uniqueResult());
			if (getCustomer() == null) {
				return;
			}
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
		if (getCustomer().getPassword() != null) {
			getCustomer().setDecryptedPassword(
					getDAOUtils().getDecrypted(
							getCustomer().getPassword(),
							(String) getSettings().getValue(
									"customerEncryptionType"),
							getAppSettings().getProperty("twoWayKeyFile")));
		}
		return;
	}

	/**
	 * Populate a collection of <code>CustomerAddress</code> objects, one for
	 * each of the addresses associated with this customer.
	 */
	public Collection loadAddressCollection() throws Exception {
		Hibernate.initialize(getCustomer().getCustomerAddressCollection());
		return getCustomer().getCustomerAddressCollection();
	}

	/**
	 * Populate a collection of <code>CustomerWishList</code> objects, one for
	 * each of the wish lists associated with this customer.
	 */
	public Collection loadCustomerWishLists() throws Exception {
		Hibernate.initialize(getCustomer().getCustomerWishLists());
		if (getCustomer().getCustomerWishLists() != null) {
			Iterator i = getCustomer().getCustomerWishLists().iterator();
			while (i.hasNext()) {
				CustomerWishList cwl = (CustomerWishList) i.next();
				Hibernate.initialize(cwl.getCustomerWishListProducts());
			}
		}
		return getCustomer().getCustomerWishLists();
	}

	/**
	 * Populate a collection of <code>Order</code> objects, one for each order
	 * in this customer's order history.
	 */
	public Collection loadOrderHistory() throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			Query q = getNamedQuery(session, "customer.selectOrderHistory");
			q.setInteger("customerID", getCustomer().getCustomerID());
			getCustomer().setOrderHistory(new LinkedHashSet(q.list()));
			return getCustomer().getOrderHistory();
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
	}

	public Collection loadDownloadFiles() throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			Query q = getNamedQuery(session, "customer.selectDownloadFiles");
			q.setInteger("customerID", getCustomer().getCustomerID());
			return new LinkedHashSet(q.list());
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
	}

	/**
	 * Return a collection of <code>Order</code> objects, one for each cart this
	 * customer has saved.
	 */
	public Collection loadSavedCarts() throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			// Get the number of days carts can be saved
			String savedCartDays = (String) getSettings().getValue(
					"savedCartDays");
			int daysAgo = 0;
			if (savedCartDays == null || savedCartDays.equals("0")) {
				daysAgo = -1000000;
			} else {
				daysAgo = -1 * Integer.parseInt(savedCartDays);
			}
			Calendar start = Calendar.getInstance();
			start.add(Calendar.DATE, daysAgo);
			// Construct the query
			Query q = getNamedQuery(session, "customer.selectSavedCarts");
			q.setString("daysAgo", formatDateTime(new java.util.Date(start
					.getTimeInMillis())));
			q.setInteger("customerID", getCustomer().getCustomerID());
			return new LinkedHashSet(q.list());
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
	}

	public void deleteCustomer() throws DataAccessException {
		deleteCustomer(false);
	}

	public void deleteCustomer(boolean commit) throws DataAccessException {
		Session session = getDaoFactory().startSession();
		try {
			setCustomer((Customer) session.load(Customer.class, new Integer(
					getCustomer().getCustomerID())));
			session.delete(getCustomer());
			if (commit) {
				getDaoFactory().commitTransaction();
			}
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
		return;
	}

	public void loadCustomerByPasswordToken() throws Exception {
		Session session = getDaoFactory().startSession();
		try {
			Query q = session.getNamedQuery("customer.loadCustomerByPasswordToken");
			Customer bean = (Customer) getCustomer();
			q.setString("token", bean.getPasswordToken());
			this.setCustomer((Customer) q.uniqueResult());
		} catch (HibernateException he) {
			if (log.isErrorEnabled())
				log.error("Hibernate error: " + he.toString());
			throw new DataAccessException(he);
		}
	}

}
